from __future__ import annotations
from abc import abstractmethod
from dataclasses import dataclass, field
import math
from typing import Generic, List, Optional, TypeVar, Union, overload

import torch
from torch import Tensor
from torch.optim import Optimizer
from torch.optim.lr_scheduler import CosineAnnealingLR, _LRScheduler

from ranzen.decorators import implements

__all__ = [
    "CosineLRWithLinearWarmup",
    "CosineWarmup",
    "ExponentialWarmup",
    "LinearWarmup",
    "LinearWarmupLR",
    "Scheduler",
    "WarmupScheduler",
]


class LinearWarmupLR(_LRScheduler):
    """
    Applies a linear warmup schedule to the learning rate, increasing/decreasing it by a fixed
    step-size from lr_start to the base value over the specified number of warmup steps.
    """

    last_epoch: int
    base_lrs: List[float]
    optimizer: Optimizer

    def __init__(
        self, optimizer: Optimizer, *, warmup_iters: int, lr_start: float = 0, last_epoch: int = -1
    ) -> None:
        self.lr_start = lr_start
        self.warmup_iters = warmup_iters
        super().__init__(optimizer=optimizer, last_epoch=last_epoch)

    @implements(_LRScheduler)
    def get_lr(self) -> list[float]:
        """
        Get the learning rate of each parameter group.

        :returns: The learning rate for each parameter group in the optimizer.
        """
        if self.last_epoch > self.warmup_iters:
            return [group["lr"] for group in self.optimizer.param_groups]

        lrs = [
            self.lr_start + self._get_step_size(base_lr) * self.last_epoch
            for group, base_lr in zip(self.optimizer.param_groups, self.base_lrs)
        ]
        return lrs

    def _get_step_size(self, base_lr: float) -> float:
        return (base_lr - self.lr_start) / self.warmup_iters

    def _get_closed_form_lr(self) -> list[float]:
        return [
            self.lr_start
            + (base_lr - self.lr_start)
            * min(self.last_epoch, self.warmup_iters)
            / self.warmup_iters
            for base_lr in self.base_lrs
        ]


class CosineLRWithLinearWarmup(_LRScheduler):
    """
    Sets the learning rate of each parameter group to follow a linear warmup schedule between
    lr_start and base_lr followed by a cosine annealing schedule between base_lr and lr_min.
    """

    last_epoch: int
    base_lrs: List[float]
    optimizer: Optimizer

    def __init__(
        self,
        optimizer: Optimizer,
        *,
        warmup_iters: Union[int, float],
        lr_start: float = 0.0,
        total_iters: int,
        lr_min: float = 0.0,
        last_epoch: int = -1,
    ) -> None:
        """
        :param optimizer: Optimizer whose parameter groups are to be scheduled.
        :param warmup_iters: Maximum number of iterations for linear warmup.
            Float values are interpreted as a fraction of ``total_iters``.

        :param total_iters: Total number of iterations.
        :param lr_start: Learning rate at the beginning of linear warmup.
        :param lr_min: Minimum learning rate permitted with cosine annealing.
        :param last_epoch: The index of the last epoch.

        :raises AttributeError: If warmup_iters is a float and not in the range [0, 1].
        """
        self.total_iters = total_iters
        self.lr_start = lr_start
        self.lr_min = lr_min
        if isinstance(warmup_iters, float):
            if not (0 <= warmup_iters <= 1):
                raise AttributeError(
                    "If 'warmup_iters' is a float, it must be in the range [0, 1]."
                )
            warmup_iters = round(warmup_iters * total_iters)
        self.warmup_iters = warmup_iters
        self._scheduler: Union[LinearWarmupLR, CosineAnnealingLR] = LinearWarmupLR(
            optimizer=optimizer, warmup_iters=warmup_iters, lr_start=lr_start
        )
        super().__init__(optimizer=optimizer, last_epoch=last_epoch)

    @property
    def scheduler(self) -> Union[LinearWarmupLR, CosineAnnealingLR]:
        """
        The scheduler currently in use, as determined by the curren step. If the current stepe
        exceeds the number of warmup iterations, then the cosine scheduler will be returned, else
        the learning-warmup scheduler will be.

        :returns: The learning-rate scheduler currently in use.
        """
        # Switch to cosine-scheduling once the warmup-period is complete.
        if self.last_epoch == self.warmup_iters:
            self._scheduler = CosineAnnealingLR(
                optimizer=self.optimizer,
                T_max=self.total_iters - self.warmup_iters + 1,
                eta_min=self.lr_min,
            )
        return self._scheduler

    def get_lr(self) -> list[float]:
        """
        Get the learning rate of each parameter group.

        :returns: The learning rate for each parameter group in the optimizer.
        """
        return self.scheduler.get_lr()  # type: ignore

    def step(self, epoch: Optional[int] = None) -> None:
        """
        Update the learning rates using the currently-used scheduler.
        """
        self.last_epoch = (self.last_epoch + 1) if epoch is None else epoch
        return self.scheduler.step(epoch)


T = TypeVar("T", Tensor, float)


@dataclass(unsafe_hash=True)
class Scheduler(Generic[T]):
    start_val: T
    val: T = field(init=False)

    def __post_init__(self) -> None:
        self.val = self.start_val

    @abstractmethod
    def _update(self, value: T) -> T:
        ...

    @torch.no_grad()
    def step(self) -> None:
        """Update the scheduled value."""
        self.val = self._update(self.val)

    @overload
    def __add__(self: "Scheduler[Tensor]", other: float) -> Tensor:
        ...

    @overload
    def __add__(self: "Scheduler[Tensor]", other: Tensor) -> Tensor:
        ...

    @overload
    def __add__(self: "Scheduler[float]", other: float) -> float:
        ...

    @overload
    def __add__(self: "Scheduler[float]", other: Tensor) -> Tensor:
        ...

    def __add__(self, other: Union[Tensor, float]) -> Union[Tensor, float]:
        return other + self.val

    @overload
    def __mul__(self: "Scheduler[Tensor]", other: float) -> Tensor:
        ...

    @overload
    def __mul__(self: "Scheduler[Tensor]", other: Tensor) -> Tensor:
        ...

    @overload
    def __mul__(self: "Scheduler[float]", other: float) -> float:
        ...

    @overload
    def __mul__(self: "Scheduler[float]", other: Tensor) -> Tensor:
        ...

    def __mul__(self, other: Union[Tensor, float]) -> Union[Tensor, float]:
        return other * self.val

    @overload
    def __imul__(self: "Scheduler[Tensor]", other: float) -> Tensor:
        ...

    @overload
    def __imul__(self: "Scheduler[Tensor]", other: Tensor) -> Tensor:
        ...

    @overload
    def __imul__(self: "Scheduler[float]", other: float) -> float:
        ...

    @overload
    def __imul__(self: "Scheduler[float]", other: Tensor) -> Tensor:
        ...

    def __imul__(self, other: Union[Tensor, float]) -> Union[Tensor, float]:
        return other * self.val

    @overload
    def __rmul__(self: "Scheduler[Tensor]", other: float) -> Tensor:
        ...

    @overload
    def __rmul__(self: "Scheduler[Tensor]", other: Tensor) -> Tensor:
        ...

    @overload
    def __rmul__(self: "Scheduler[float]", other: float) -> float:
        ...

    @overload
    def __rmul__(self: "Scheduler[float]", other: Tensor) -> Tensor:
        ...

    def __rmul__(self, other: Union[Tensor, float]) -> Union[Tensor, float]:
        return other * self.val


@dataclass(unsafe_hash=True)
class WarmupScheduler(Scheduler[T]):
    end_val: T
    warmup_steps: int
    _curr_step: int = field(init=False)

    @implements(Scheduler)
    def __post_init__(self) -> None:
        if self.warmup_steps < 0:
            raise AttributeError("'warmup_steps' must be a non-negative integer.")
        super().__post_init__()
        if self.warmup_steps == 0:
            self.val = self.end_val
        self._curr_step = 0

    @property
    def warmed_up(self) -> bool:
        return self._curr_step == self.warmup_steps

    @torch.no_grad()
    @implements(Scheduler)
    def step(self) -> None:
        if not self.warmed_up:
            super().step()
            self._curr_step += 1


@dataclass(unsafe_hash=True)
class LinearWarmup(WarmupScheduler[T]):
    step_size: T = field(init=False)

    @implements(WarmupScheduler)
    def __post_init__(self) -> None:
        super().__post_init__()
        if self.warmup_steps == 0:
            self.step_size = 0
        else:
            self.step_size = (self.end_val - self.start_val) / self.warmup_steps

    @implements(WarmupScheduler)
    def _update(self, value: T) -> T:
        return value + self.step_size


@dataclass(unsafe_hash=True)
class ExponentialWarmup(WarmupScheduler[T]):
    end_val: T
    step_size: T = field(init=False)

    @implements(WarmupScheduler)
    def __post_init__(self) -> None:
        super().__post_init__()
        if self.warmup_steps == 0:
            self.step_size = 0
        else:
            self.step_size = (self.end_val / self.start_val) ** (1 / self.warmup_steps)

    @implements(WarmupScheduler)
    def _update(self, value: T) -> T:
        return value * self.step_size


@dataclass(unsafe_hash=True)
class CosineWarmup(WarmupScheduler[T]):
    end_val: T
    _coeff: T = field(init=False)

    @implements(WarmupScheduler)
    def __post_init__(self) -> None:
        super().__post_init__()
        self._coeff = 0.5 * (self.end_val - self.start_val)

    @implements(WarmupScheduler)
    def _update(self, value: T) -> T:
        if self.warmup_steps == 0:
            return value
        return (
            self._coeff * (1 - math.cos(math.pi * (self._curr_step + 1) / self.warmup_steps))
            + self.start_val
        )



from typing import TypeVar

from typing_extensions import Protocol, Self, runtime_checkable

__all__ = ["Sized", "Addable"]


T_co = TypeVar("T_co", covariant=True)


@runtime_checkable
class Sized(Protocol[T_co]):
    def __len__(self) -> int:
        ...


@runtime_checkable
class Addable(Protocol[T_co]):
    def __add__(self, other: Self, /) -> Self:
        ...



from __future__ import annotations
from typing import Sequence

import torch
from torch import Tensor

__all__ = ["batched_randint"]


def batched_randint(
    high: Tensor,
    *,
    size: int | Sequence[int] | None = None,
    generator: torch.Generator | None = None,
) -> Tensor:
    r"""Batched version of :func:`torch.randint`.

    Randomly samples an integer from the domain :math:`[0, h_i]` for each sample :math:`h_i \in high`.

    :func:`torch.randint` requires ``high`` to be an integer and thus  prohibits having
    different samples within a batch having sampling domains, something which is necessary in order
    to vectorise, for example, sampling from groups of different sizes or sampling objects with
    different offsets. This func addresses this limitation using inverse transform sampling.

    :param high: A batch of tensors encoding the maximum integer value the corresponding random
        samples may take.

    :param size: An integer or sequence of integers defining the shape of the output tensor for each
        upper-bound specified in ``high``. The overall size of the sampled tensor will be
        'size(``high``) + ``size``. If ``None`` then the output size is simply 'size(``high``)'.

    :param generator: Pseudo-random-number generator to use for sampling.

    :returns: A tensor of random-sampled integers upper-bounded by the values in ``high``.
    """
    total_size = high.size()
    if size is not None:
        total_size = list(total_size)
        if isinstance(size, int):
            total_size.append(size)
        else:
            total_size.extend(size)

    step_sizes = high.reciprocal()
    u = (
        torch.rand(size=total_size, device=high.device, generator=generator) * (1 + step_sizes)
        - step_sizes / 2
    )
    return (u.clamp(min=0, max=1) * (high - 1)).round().long()



from __future__ import annotations
from collections.abc import Iterable, Iterator
from datetime import datetime
import random
from typing import Any, List, TypeVar

import numpy as np
import torch
from torch import Tensor
import torch.nn as nn

__all__ = ["count_parameters", "random_seed", "inf_generator", "Event", "batchwise_pdist"]


def count_parameters(model: nn.Module) -> int:
    """Count all parameters (that have a gradient) in the given model"""
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def random_seed(seed_value: int, *, use_cuda: bool) -> None:
    np.random.seed(seed_value)  # cpu vars
    torch.manual_seed(seed_value)  # cpu  vars
    random.seed(seed_value)  # Python
    if use_cuda:
        torch.cuda.manual_seed(seed_value)  # type: ignore
        torch.cuda.manual_seed_all(seed_value)  # type: ignore
        torch.backends.cudnn.deterministic = True  # type: ignore
        torch.backends.cudnn.benchmark = False  # type: ignore


T = TypeVar("T")


def inf_generator(iterable: Iterable[T]) -> Iterator[T]:
    """Get DataLoaders in a single infinite loop.

    for i, (x, y) in enumerate(inf_generator(train_loader))
    """
    iterator = iter(iterable)
    # try to take one element to ensure that the iterator is not empty
    first_value = next(iterator, None)
    if first_value is not None:
        yield first_value
    else:
        raise RuntimeError("The given iterable is empty.")
    while True:
        try:
            yield next(iterator)
        except StopIteration:
            iterator = iter(iterable)


class Event:
    """Emulates torch.cuda.Event, but supports running on a CPU too.

    :example:
        >>> from ranzen.torch import Event
        >>> with Event() as event:
        >>>     y = some_nn_module(x)
        >>> print(event.time)
    """

    def __init__(self):
        self.time = 0.0
        self._cuda = torch.cuda.is_available()  # type: ignore
        self._event_start: torch.cuda.Event | datetime  # type: ignore

    def __enter__(self) -> Event:
        """Mark a time.

        Mimics torch.cuda.Event.
        """
        if self._cuda:
            self._event_start = torch.cuda.Event(enable_timing=True)  # type: ignore
            self._event_start.record()  # type: ignore
        else:
            self._event_start = datetime.now()
        return self

    def __exit__(self, *args: Any) -> None:
        if self._cuda:
            event_end = torch.cuda.Event(enable_timing=True)  # type: ignore
            event_end.record(stream=torch.cuda.current_stream())
            torch.cuda.synchronize()  # type: ignore
            assert isinstance(self._event_start, torch.cuda.Event)  # type: ignore
            self.time = self._event_start.elapsed_time(event_end)  # type: ignore
        else:
            assert isinstance(self._event_start, datetime)
            self.time = datetime.now().microsecond - self._event_start.microsecond

    def __repr__(self) -> str:
        return f"Event of duration: {self.time}"


def batchwise_pdist(x: Tensor, chunk_size: int = 1000, p_norm: float = 2.0) -> Tensor:
    """Compute pair-wise distance in batches.

    This is sometimes necessary because if you compute pdist directly, it doesn't fit into memory.

    :param x: Tensor of shape (N, F) where F is the number of features.
    :param chunk_size: Size of the chunks that are used to compute the pair-wise distance. Larger
        chunk size is probably faster, but may not fit into memory.
    :param p_norm: Which norm to use for the distance. Euclidean norm by default.
    :returns: All pair-wise distances in a tensor of shape (N, N).
    """
    chunks = torch.split(x, chunk_size)

    columns: List[Tensor] = []
    for chunk in chunks:
        shards = [torch.cdist(chunk, other_chunk, p_norm) for other_chunk in chunks]
        column = torch.cat(shards, dim=1)
        # the result has to be moved to the CPU; otherwise we'll run out of GPU memory
        columns.append(column.cpu())

        # free up memory
        for shard in shards:
            del shard
        del column
        torch.cuda.empty_cache()

    dist = torch.cat(columns, dim=0)

    # free up memory
    for column in columns:
        del column
    torch.cuda.empty_cache()
    return dist



from __future__ import annotations
import copy
from enum import Enum
import functools
import operator
import sys
from typing import Any, Dict, Iterable, MutableMapping, Type, TypeVar, overload

from typing_extensions import Self

from ranzen.types import Addable

__all__ = [
    "AddDict",
    "StrEnum",
    "flatten_dict",
    "gcopy",
    "reduce_add",
    "str_to_enum",
]


def flatten_dict(
    d: MutableMapping[str, Any], *, parent_key: str = "", sep: str = "."
) -> dict[str, Any]:
    """Flatten a nested dictionary by separating the keys with `sep`.

    :param d: Dictionary to be flattened.
    :param parent_key: Key-prefix (separated from the key with 'sep') to use for top-level
        keys of the flattened dictionary.
    :param sep: Character to separate the parent keys from the child keys with at each level with.

    :returns: Flattened dictionary with keys capturing the nesting path as ``parent_key.child_key``,
        where 'parent_key' is defined recursively, with base value 'parent_key' as specified in the
        function call.
    """
    items: list[tuple[str, Any]] = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, parent_key=new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


T = TypeVar("T")


@overload
def gcopy(obj: T, *, deep: bool = True, num_copies: None = ..., **kwargs: Any) -> T:
    ...


@overload
def gcopy(obj: T, *, deep: bool = True, num_copies: int, **kwargs: Any) -> list[T]:
    ...


def gcopy(
    obj: T, *, deep: bool = True, num_copies: int | None = None, **kwargs: Any
) -> T | list[T]:
    """Generalised (g) copy function.
    Allows for switching between deep and shallow copying within a single function
    as well as for the creation of multiple copies and for copying while simultaneously
    attribute-setting.

    :param obj: Object to be copied.
    :param deep: Whether to create deep (True) or shallow (False) copies.
    :param num_copies: Number of copies to create with 'None' being equivalent to 1.
    :param kwargs: Key-word arguments specifying a name of an attribute and the
        new value to set it to in the copies.

    :returns: A copy or list of copies (if num_copies > 1) of the object 'obj'.

    :raises AttributeError: If an attribute specified in ``kwargs`` doesn't exist.
    """
    if num_copies is not None:
        return [gcopy(obj=obj, deep=deep, num_copies=None, **kwargs) for _ in range(num_copies)]
    copy_fn = copy.deepcopy if deep else copy.copy
    obj_cp = copy_fn(obj)
    for attr, value in kwargs.items():
        if not hasattr(obj_cp, attr):
            raise AttributeError(
                f"Object of type '{type(obj_cp).__name__}' has no attribute '{attr}'."
            )
        setattr(obj_cp, attr, value)
    return obj_cp


E = TypeVar("E", bound=Enum)


def str_to_enum(str_: str | E, *, enum: type[E]) -> E:
    """Convert a string to an enum based on name instead of value.
    If the string is not a valid name of a member of the target enum,
    an error will be raised.

    :param str_: String to be converted to an enum member of type ``enum``.
    :param enum: Enum class to convert ``str_`` to.

    :returns: The enum member of type ``enum`` with name ``str_``.

    :raises TypeError: if the given string is not a valid name of a member of the target enum
    """
    if isinstance(str_, enum):
        return str_
    try:
        return enum[str_]  # type: ignore
    except KeyError:
        valid_ls = [mem.name for mem in enum]
        raise TypeError(
            f"'{str_}' is not a valid option for enum '{enum.__name__}'; must be one of {valid_ls}."
        )


if sys.version_info >= (3, 11):
    # will be available in python 3.11
    from enum import StrEnum  # type: ignore
else:
    #
    # the following is copied straight from https://github.com/python/cpython/blob/3.11/Lib/enum.py
    #
    # DO NOT CHANGE THIS CODE!
    #
    from enum import Enum

    class ReprEnum(Enum):
        """
        Only changes the repr(), leaving str() and format() to the mixed-in type.
        """

    _S = TypeVar("_S", bound="StrEnum")

    class StrEnum(str, ReprEnum):
        """
        Enum where members are also (and must be) strings
        """

        _value_: str

        def __new__(cls: Type[_S], *values: str) -> _S:
            "values must already be of type `str`"
            if len(values) > 3:
                raise TypeError("too many arguments for str(): %r" % (values,))
            if len(values) == 1:
                # it must be a string
                if not isinstance(values[0], str):
                    raise TypeError("%r is not a string" % (values[0],))
            if len(values) >= 2:
                # check that encoding argument is a string
                if not isinstance(values[1], str):
                    raise TypeError("encoding must be a string, not %r" % (values[1],))
            if len(values) == 3:
                # check that errors argument is a string
                if not isinstance(values[2], str):
                    raise TypeError("errors must be a string, not %r" % (values[2]))
            value = str(*values)
            member = str.__new__(cls, value)
            member._value_ = value
            return member

        __str__ = str.__str__  # type: ignore

        def _generate_next_value_(name: str, start: int, count: int, last_values: list[str]) -> str:
            """
            Return the lower-cased version of the member name.
            """
            return name.lower()


_KT = TypeVar("_KT")
_VT = TypeVar("_VT", bound=Addable)
_VT2 = TypeVar("_VT2", bound=Addable)


class AddDict(Dict[_KT, _VT], Addable):
    """
    Extension of the built-in dictionary class that supports the use of the ``__add__`` operator for
    key-wise addition.

    :example:
        .. code-block:: python

            # Simple case of addition of integers.
            d1 = AddDict({"foo": 1, "bar": 2})
            d2 = {"foo": 3, "bar": 4}
            d1 + d2 # {'foo': 4, 'bar': 6}

            # Concatenation of lists
            d3 = AddDict({"foo": [1], "bar": [2]})
            d4 = {"foo": [3, 4], "bar": 4}
            d3 + d4 # {'foo': [1, 3, 4], 'bar': [2, 4]}
    """

    @overload
    def __add__(
        self: Self,
        other: int,
    ) -> Self:
        ...

    @overload
    def __add__(
        self: Self,
        other: dict[_KT, _VT2],
    ) -> AddDict[_KT, _VT | _VT2]:
        ...

    def __add__(
        self: Self,
        other: int | dict[_KT, _VT2],
    ) -> Self | AddDict[_KT, _VT | _VT2]:
        # Allow ``other`` to be an integer, but specifying the identity function, for compatibility
        # with th 'no-default' version of``sum``.
        if isinstance(other, int):
            return self
        copy = AddDict()
        copy.update(gcopy(self, deep=False))

        for key_o, value_o in other.items():
            if not isinstance(value_o, Addable):
                raise TypeError(f"Value of type '{type(value_o)}' is not addable.")
            if key_o in self:
                value_s = self[key_o]
                if not isinstance(value_s, Addable):
                    raise TypeError(f"Value of type '{type(value_s)}' is not addable.")
                try:
                    # Values are mutually addable (but not necessarily of the same type).
                    copy[key_o] = value_s + value_o
                except TypeError as e:
                    msg = (
                        f"Values of type '{type(value_s)}' and '{type(value_o)}' for key "
                        f"'{key_o}' are not mutuablly addable."
                    )
                    raise TypeError(msg) from e
            else:
                copy[key_o] = value_o
        return copy

    @overload
    def __radd__(
        self: Self,
        other: int,
    ) -> Self:
        ...

    @overload
    def __radd__(
        self: Self,
        other: dict[_KT, _VT2],
    ) -> AddDict[_KT, _VT | _VT2]:
        ...

    def __radd__(self: Self, other: int | dict[_KT, _VT2]) -> Self | AddDict[_KT, _VT | _VT2]:
        return self + other


A = TypeVar("A", bound=Addable)


def reduce_add(sequence: Iterable[A]) -> A:
    """
    Sum an iterable using functools.reduce and operator.add rather than the built-in sum operator
    to bypass need to specify an initial value or have the elements ``__add__`` operator be
    compatible with integers (0 being the default initial value).

    :param sequence: An iterable of addable instances, all of the same type (invariant).
    :returns: The sum of all elements in ``__iterable``.
    """
    return functools.reduce(operator.add, sequence)



from __future__ import annotations
from dataclasses import dataclass
from typing import Any

import torch.nn as nn
from typing_extensions import Self, final

__all__ = ["DcModule"]


@dataclass(unsafe_hash=True)
class DcModule(nn.Module):
    @final
    def __new__(cls: type[Self], *args: Any, **kwargs: Any) -> Self:
        obj = object.__new__(cls)
        nn.Module.__init__(obj)
        return obj



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ranzen',
 'ranzen.hydra',
 'ranzen.torch',
 'ranzen.torch.optimizers',
 'ranzen.torch.transforms']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'ranzen',
    'version': '2.0.5',
    'description': 'A toolkit facilitating machine-learning experimentation.',
    'long_description': '# ranzen 🎒\n\nA python toolkit facilitating machine-learning experimentation.\n\n[Documentation](https://wearepal.github.io/ranzen/)\n\n## Install\n\nRun\n```\npip install ranzen\n```\n\nor install directly from GitHub:\n```\npip install git+https://github.com/predictive-analytics-lab/ranzen.git@main\n```\n',
    'author': 'PAL',
    'author_email': 'info@predictive-analytics-lab.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/predictive-analytics-lab/ranzen',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8.0,<3.11',
}


setup(**setup_kwargs)



from __future__ import annotations
from typing import Optional, cast, overload

import torch
from torch import Tensor
import torch.distributions as td
import torch.nn.functional as F

from ranzen.torch.sampling import batched_randint
from ranzen.torch.transforms.mixup import InputsTargetsPair

__all__ = ["RandomCutMix"]


class RandomCutMix:
    r"""Randomly apply CutMix to a batch of images.
    PyTorch implementation of the the `CutMix`_ image-augmentation strategy.

    This implementation samples the bounding-box coordinates independently for each pair of samples
    being mixed, and, unlike other implementations, does so in a way that is fully-vectorised.

    .. _CutMix:
        https://arxiv.org/abs/1905.04899

    .. note::
        This implementation randomly mixes images within a batch.

    """

    def __init__(
        self,
        alpha: float = 1.0,
        *,
        p: float = 0.5,
        num_classes: int | None = None,
        inplace: bool = False,
        seed: Optional[int] = None,
    ) -> None:
        """
        :param alpha: hyperparameter of the Beta distribution used for sampling the areas
            of the bounding boxes.

        :param num_classes: The total number of classes in the dataset that needs to be specified if
            wanting to mix up targets that are label-enoded. Passing label-encoded targets without
            specifying ``num_classes`` will result in a RuntimeError.

        :param p: The probability with which the transform will be applied to a given sample.

        :param inplace: Whether the transform should be performed in-place.

        :param seed: The PRNG seed to use for sampling pairs and bounding-box coordinates.

        :raises ValueError: if ``p`` is not in the range [0, 1] , if ``num_classes < 1``, or if
            ``alpha`` is not a positive real number.

        """
        super().__init__()
        if not 0 <= p <= 1:
            raise ValueError("'p' must be in the range [0, 1].")
        self.p = p
        if alpha < 0:
            raise ValueError("'alpha' must be positive.")
        self.alpha = alpha
        if (num_classes is not None) and num_classes < 1:
            raise ValueError(f"{ num_classes } must be greater than 1.")
        self.lambda_sampler = td.Beta(concentration0=alpha, concentration1=alpha)
        self.num_classes = num_classes
        self.inplace = inplace
        self.seed = seed

    def _sample_masks(
        self,
        inputs: Tensor,
        *,
        num_samples: int,
        generator: torch.Generator | None = None,
    ) -> tuple[Tensor, Tensor]:
        height, width = inputs.shape[-2:]
        lambdas: Tensor = self.lambda_sampler.sample(sample_shape=torch.Size((num_samples,))).to(
            inputs.device
        )
        side_props = torch.sqrt(1.0 - lambdas)
        box_heights = (side_props * height).round()
        box_widths = (side_props * width).round()
        box_coords_y1 = batched_randint(height - box_heights, generator=generator)
        box_coords_x1 = batched_randint(width - box_widths, generator=generator)
        # Compute the terminal y-coördinates for the bounding boxes
        box_coords_y2 = box_coords_y1 + box_heights
        box_coords_x2 = box_coords_x1 + box_widths
        # Convert the bounding box coordinates into masks.
        y_indices = torch.arange(height, device=inputs.device).unsqueeze(0).expand(num_samples, -1)
        y_masks = (box_coords_y2.unsqueeze(-1) > y_indices) & (
            y_indices >= box_coords_y1.unsqueeze(-1)
        )
        x_indices = torch.arange(width, device=inputs.device).unsqueeze(0).expand(num_samples, -1)
        x_masks = (box_coords_x2.unsqueeze(-1) > x_indices) & (
            x_indices >= box_coords_x1.unsqueeze(-1)
        )
        masks = (
            (y_masks.unsqueeze(1) * x_masks.unsqueeze(-1))
            .unsqueeze(1)
            .expand(-1, inputs.size(1), -1, -1)
        )
        cropped_area_ratios = (box_widths * box_widths) / (width * height)

        return masks, cropped_area_ratios

    @overload
    def _transform(self, inputs: Tensor, *, targets: Tensor) -> InputsTargetsPair:
        ...

    @overload
    def _transform(self, inputs: Tensor, *, targets: None = ...) -> Tensor:
        ...

    def _transform(
        self, inputs: Tensor, *, targets: Tensor | None = None
    ) -> Tensor | InputsTargetsPair:
        if inputs.ndim != 4:
            raise ValueError(f"'inputs' must be a batch of image tensors of shape (C, H, W).")
        batch_size = len(inputs)
        if (targets is not None) and (batch_size != len(targets)):
            raise ValueError(f"'inputs' and 'targets' must match in size at dimension 0.")

        generator = (
            None if self.seed is None else torch.Generator(inputs.device).manual_seed(self.seed)
        )
        if (batch_size == 1) or (self.p == 0):
            return inputs if targets is None else InputsTargetsPair(inputs=inputs, targets=targets)
        elif self.p < 1:
            # Sample a mask determining which samples in the batch are to be transformed
            selected = torch.rand(batch_size, device=inputs.device, generator=generator) < self.p
            num_selected = int(selected.count_nonzero())
            if num_selected == 0:
                return (
                    inputs if targets is None else InputsTargetsPair(inputs=inputs, targets=targets)
                )
            indices = selected.nonzero(as_tuple=False).long().flatten()
        # If p >= 1 then the transform is always applied and we can skip the sampling step above.
        else:
            num_selected = batch_size
            indices = torch.arange(batch_size, device=inputs.device, dtype=torch.long)

        # Pair each selected sample with another sample that will serve as the 'patch donor'
        pair_indices = torch.arange(num_selected).roll(1, 0)
        masks, cropped_area_ratios = self._sample_masks(
            inputs=inputs, num_samples=num_selected, generator=generator
        )

        if not self.inplace:
            inputs = inputs.clone()
        # Trnasplant patches from the paired images to the anchor images as determined by the masks.
        inputs[indices] = ~masks * inputs[indices] + masks * inputs[pair_indices]
        # No targets were recevied so we're done.
        if targets is None:
            return inputs

        # Targets are label-encoded and need to be one-hot encoded prior to mixup.
        if torch.atleast_1d(targets.squeeze()).ndim == 1:
            if self.num_classes is None:
                raise RuntimeError(
                    f"{self.__class__.__name__} can only be applied to label-encoded targets if "
                    "'num_classes' is specified."
                )
            targets = cast(Tensor, F.one_hot(targets, num_classes=self.num_classes))
        elif not self.inplace:
            targets = targets.clone()
        # Targets need to be floats to be mixed up.
        targets = targets.float()
        target_lambdas = 1.0 - cropped_area_ratios
        target_lambdas.unsqueeze_(-1)
        targets[indices] *= target_lambdas
        targets[indices] += (1.0 - target_lambdas) * targets[pair_indices]

        return InputsTargetsPair(inputs=inputs, targets=targets)

    @overload
    def __call__(self, inputs: Tensor, *, targets: Tensor) -> InputsTargetsPair:
        ...

    @overload
    def __call__(self, inputs: Tensor, *, targets: None = ...) -> Tensor:
        ...

    def __call__(
        self, inputs: Tensor, *, targets: Tensor | None = None
    ) -> Tensor | InputsTargetsPair:
        """
        :param inputs: The samples to apply mixup to.
        :param targets: The corresponding targets to apply mixup to. If the targets are
            label-encoded then the 'num_classes' attribute cannot be None.

        :return: If target is None, the Tensor of cutmix-transformed inputs. If target is not None, a
            namedtuple containing the Tensor of cutmix-transformed inputs (inputs) and the
            corresponding Tensor of cutmix-transformed targets (targets).
        """
        return self._transform(inputs=inputs, targets=targets)



from __future__ import annotations
from enum import Enum, auto
from functools import partial
from typing import Optional, Union

from torch import Tensor, nn
import torch.nn.functional as F

from ranzen import parsable, str_to_enum

__all__ = [
    "CrossEntropyLoss",
    "ReductionType",
    "cross_entropy_loss",
    "reduce",
]


class ReductionType(Enum):
    """An enum for the type of reduction to apply to a batch of losses."""

    mean = auto()
    """compute the mean over all dimensions."""
    none = auto()
    """no reduction."""
    sum = auto()
    """compute the sum over all dimensions."""
    batch_mean = auto()
    """compute the mean over the batch (first) dimension, the sum over the remaining dimensions."""


def reduce(losses: Tensor, reduction_type: ReductionType | str) -> Tensor:
    if isinstance(reduction_type, str):
        reduction_type = str_to_enum(str_=reduction_type, enum=ReductionType)
    if reduction_type is ReductionType.mean:
        return losses.mean()
    elif reduction_type is ReductionType.batch_mean:
        return losses.sum() / losses.size(0)
    elif reduction_type is ReductionType.sum:
        return losses.sum()
    elif reduction_type is ReductionType.none:
        return losses
    raise TypeError(
        f"Received invalid type '{type(reduction_type)}' for argument 'reduction_type'."
    )


def cross_entropy_loss(
    input: Tensor,
    *,
    target: Tensor,
    instance_weight: Tensor | None = None,
    reduction: ReductionType | str = ReductionType.mean,
    ignore_index: int = -100,
    class_weight: Tensor | None = None,
    label_smoothing: float = 0.0,
) -> Tensor:
    r"""This criterion computes the cross entropy loss between input and target.

    See :class:`~CrossEntropyLoss` for details.

    :param input: Predicted unnormalized scores (often referred to as logits).
    :param target: Ground truth class indices or class probabilities.

    :param instance_weight: a manual rescaling weight given to each sample. If given, has to be a
        Tensor of 'N'.

    :param class_weight: A manual rescaling weight given to each class. If given, has to be a
        Tensor of size `C`.

    :param ignore_index: Specifies a target value that is ignored and does not contribute to the
        input gradient. Note that :attr:`ignore_index` is only applicable when the target contains
        class indices.

    :param reduction: Specifies the reduction to apply to the output.

    :param label_smoothing: A float in [0.0, 1.0]. Specifies the amount of smoothing when computing
        the loss, where 0.0 means no smoothing. The targets become a mixture of the original ground
        truth and a uniform distribution as described in `Rethinking the Inception Architecture for
        Computer Vision <https://arxiv.org/abs/1512.00567>`__. Default: :math:`0.0`.

    :returns: The (reduced) cross-entropy between ``input`` and ``target``.

    :example:
        >>> # Example of target with class indices
        >>> input = torch.randn(3, 5, requires_grad=True)
        >>> target = torch.randint(5, (3,), dtype=torch.int64)
        >>> loss = F.cross_entropy(input, target)
        >>> loss.backward()
        >>>
        >>> # Example of target with class probabilities
        >>> input = torch.randn(3, 5, requires_grad=True)
        >>> target = torch.randn(3, 5).softmax(dim=1)
        >>> loss = F.cross_entropy(input, target)
        >>> loss.backward()
    """
    if isinstance(reduction, str):
        reduction = str_to_enum(str_=reduction, enum=ReductionType)
    if input.ndim == 1 or input.size(1) == 1:  # Binary classification
        target = target.view_as(input)
        if not target.is_floating_point():
            target = target.float()
        loss_fn = F.binary_cross_entropy_with_logits
    else:  # Multiclass classification
        target = target.view(input.size(0), -1).squeeze(-1)
        if (target.ndim == 1) and target.is_floating_point():
            target = target.long()
        loss_fn = partial(F.cross_entropy, ignore_index=ignore_index)
    losses = loss_fn(
        input=input,
        target=target,
        weight=class_weight,
        reduction="none",
    )
    if instance_weight is not None:
        losses *= instance_weight.view_as(losses)
    return reduce(losses=losses, reduction_type=reduction)


class CrossEntropyLoss(nn.Module):
    weight: Tensor | None

    @parsable
    def __init__(
        self,
        *,
        class_weight: Optional[Tensor] = None,
        ignore_index: int = -100,
        reduction: Union[ReductionType, str] = ReductionType.mean,
        label_smoothing: float = 0.0,
    ) -> None:
        r"""This criterion computes the cross entropy loss between input and target.

        It is useful when training a classification problem with `C` classes.
        If provided, the optional argument :attr:`weight` should be a 1D `Tensor`
        assigning weight to each of the classes.
        This is particularly useful when you have an unbalanced training set.

        The `input` is expected to contain raw, unnormalized scores for each class.
        `input` has to be a Tensor of size :math:`(C)` for unbatched input,
        :math:`(minibatch, C)` or :math:`(minibatch, C, d_1, d_2, ..., d_K)` with :math:`K \geq 1` for the
        `K`-dimensional case. The last being useful for higher dimension inputs, such
        as computing cross entropy loss per-pixel for 2D images.

        The `target` that this criterion expects should contain either:

        - Class indices in the range :math:`[0, C)` where :math:`C` is the number of classes; if
          `ignore_index` is specified, this loss also accepts this class index (this index
          may not necessarily be in the class range). The unreduced (i.e. with :attr:`reduction`
          set to ``'none'``) loss for this case can be described as:

          .. math::
              \ell(x, y) = L = \{l_1,\dots,l_N\}^\top, \quad
              l_n = - w_{y_n} \log \frac{\exp(x_{n,y_n})}{\sum_{c=1}^C \exp(x_{n,c})}
              \cdot \mathbb{1}\{y_n \not= \text{ignore\_index}\}

          where :math:`x` is the input, :math:`y` is the target, :math:`w` is the weight,
          :math:`C` is the number of classes, and :math:`N` spans the minibatch dimension as well as
          :math:`d_1, ..., d_k` for the `K`-dimensional case. If
          :attr:`reduction` is not ``'none'`` (default ``'mean'``), then

          .. math::
              \ell(x, y) = \begin{cases}
                  \sum_{n=1}^N \frac{1}{\sum_{n=1}^N w_{y_n} \cdot \mathbb{1}\{y_n \not= \text{ignore\_index}\}} l_n, &
                   \text{if reduction} = \text{`mean';}\\
                    \sum_{n=1}^N l_n,  &
                    \text{if reduction} = \text{`sum'.}
                \end{cases}

          Note that this case is equivalent to the combination of :class:`torch.nn.LogSoftmax` and
          :class:`torch.nn.NLLLoss`.

        - Probabilities for each class; useful when labels beyond a single class per minibatch item
          are required, such as for blended labels, label smoothing, etc. The unreduced (i.e. with
          :attr:`reduction` set to ``'none'``) loss for this case can be described as:

          .. math::
              \ell(x, y) = L = \{l_1,\dots,l_N\}^\top, \quad
              l_n = - \sum_{c=1}^C w_c \log \frac{\exp(x_{n,c})}{\sum_{i=1}^C \exp(x_{n,i})} y_{n,c}

          where :math:`x` is the input, :math:`y` is the target, :math:`w` is the weight,
          :math:`C` is the number of classes, and :math:`N` spans the minibatch dimension as well as
          :math:`d_1, ..., d_k` for the `K`-dimensional case. If
          :attr:`reduction` is not ``'none'`` (default ``'mean'``), then

          .. math::
              \ell(x, y) = \begin{cases}
                  \frac{\sum_{n=1}^N l_n}{N}, &
                   \text{if reduction} = \text{`mean';}\\
                    \sum_{n=1}^N l_n,  &
                    \text{if reduction} = \text{`sum'.}
                \end{cases}

        .. note::
            The performance of this criterion is generally better when `target` contains class
            indices, as this allows for optimized computation. Consider providing `target` as
            class probabilities only when a single class label per minibatch item is too restrictive.

    
        :param class_weight: A manual rescaling weight given to each class. If given, has to be a 
            Tensor of size `C`.

        :param ignore_index: Specifies a target value that is ignored and does not contribute to the 
            input gradient. Note that :attr:`ignore_index` is only applicable when the target contains
            class indices.

        :param reduction: Specifies the reduction to apply to the output.

        :param label_smoothing: A float in [0.0, 1.0]. Specifies the amount of smoothing when computing
            the loss, where 0.0 means no smoothing. The targets become a mixture of the original ground
            truth and a uniform distribution as described in `Rethinking the Inception Architecture for
            Computer Vision <https://arxiv.org/abs/1512.00567>`__. Default: :math:`0.0`.

        :example:
            >>> # Example of target with class indices
            >>> loss = CrossEntropyLoss()
            >>> input = torch.randn(3, 5, requires_grad=True)
            >>> target = torch.empty(3, dtype=torch.long).random_(5)
            >>> output = loss(input, target)
            >>> output.backward()
            >>>
            >>> # Example of target with class probabilities
            >>> input = torch.randn(3, 5, requires_grad=True)
            >>> target = torch.randn(3, 5).softmax(dim=1)
            >>> output = loss(input, target)
            >>> output.backward()
        """
        super().__init__()
        if isinstance(reduction, str):
            reduction = str_to_enum(str_=reduction, enum=ReductionType)
        self.register_buffer("weight", class_weight)
        self.ignore_index = ignore_index
        self.label_smoothing = label_smoothing
        self._reduction = reduction

    @property
    def reduction(self) -> ReductionType:
        return self._reduction

    @reduction.setter
    def reduction(self, value: ReductionType | str) -> None:  # type: ignore
        if isinstance(value, str):
            value = str_to_enum(str_=value, enum=ReductionType)
        self._reduction = value

    def forward(
        self,
        input: Tensor,
        *,
        target: Tensor,
        instance_weight: Tensor | None = None,
        reduction: ReductionType | str | None = None,
    ) -> Tensor:
        """
        Computes the cross entropy loss between ``input`` and ``target``.

        :param input: Predicted unnormalized scores (often referred to as logits).
        :param target: Ground truth class indices or class probabilities.

        :param instance_weight: a manual rescaling weight given to each sample. If given, has to be a
            Tensor of 'N'.

        :param reduction: Overrides :attr:`reduction`.

        :returns: The (reduced) cross-entropy between ``input`` and ``target``.
        """

        reduction = self.reduction if reduction is None else reduction
        return cross_entropy_loss(
            input=input,
            target=target,
            instance_weight=instance_weight,
            reduction=reduction,
            class_weight=self.weight,
            label_smoothing=self.label_smoothing,
        )
